from enum import Enum
from typing import List, Union


_ANSI_COLORS = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "reset": 39,
    "bright_black": 90,
    "bright_red": 91,
    "bright_green": 92,
    "bright_yellow": 93,
    "bright_blue": 94,
    "bright_magenta": 95,
    "bright_cyan": 96,
    "bright_white": 97,
}  # source: https://github.com/pallets/click/termui.py


class RGBColor(tuple, Enum):
    RED = (255, 0, 0)
    CRIMSON = (220, 20, 60)
    DEEP_PINK = (255, 20, 147)
    ORANGE_RED = (255, 69, 0)
    ORANGE = (255, 165, 0)
    GREEN = (0, 128, 0)
    LIME = (0, 255, 0)
    TEAL = (0, 128, 128)
    AQUAMARINE = (102, 205, 170)
    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)
    GOLDEN = (218, 165, 32)
    CYAN = (0, 255, 255)
    YELLOW = (255, 255, 0)


class click:
    """
    This class was made to simulate `click` (see: https://github.com/pallets/click)
    package coloring functionality.

    My goal was to keep the package fully built-in, so if you do not have this
    package installed, there will be no problem with using the `style()` method.

    However, if it is installed, do not worry - this class is only used in `logs.ColorFormatter`
    and is imported only when real `click` is not installed.
    """

    @classmethod
    def style(cls, text: str, fg: Union[int, RGBColor, str], **style_kwargs) -> str:
        """
        Style a text with ANSI color codes.

        Args:
            text (str): The text to be styled.
            fg (Union[int, RGBColor, str]): The foreground color. Can be an integer,
                `RGBColor`, or a string representing the color.
            **style_kwargs: Options from the original `click.style` method.

        Returns:
            str: The styled text with ANSI color codes.
        """
        bits: List[str] = [text]
        if isinstance(fg, int):
            _color = f"{38};5;{fg:d}"
        elif isinstance(fg, (tuple, list)):
            r, g, b = fg
            _color = f"{38};2;{r:d};{g:d};{b:d}"
        else:
            _color = str(_ANSI_COLORS[fg])
        bits.insert(0, f"\033[{_color}m")
        bits.append("\033[0m")  # ANSI reset all, to stop color from being applied to all following logs

        return "".join(bits)



import sys
from typing import TYPE_CHECKING

from pocketwelt.imports import LazyModule

__all__ = [
    "b64_decode",
    "b64_encode",
    "compress_file",
    "decompress_file",
    "fetch_module",
    "getCustomLogger",
    "hash_file",
    "load_pickle",
    "save_pickle",
    "get_parent_path",
    "run_process",
    "build_all_paths",
    "build_top_paths",
    "get_all_attributes",
    "list_all_paths",
    "stdout_to_logger",
    "BaseConfig",
    "CustomLogger",
    "ColorFormatter",
]

_import_structure = {
    "configs": ["BaseConfig"],
    "encodings": ["b64_decode", "b64_encode"],
    "files": ["compress_file", "decompress_file"],
    "logs": ["getCustomLogger", "stdout_to_logger", "CustomLogger", "ColorFormatter"],
    "modules": ["fetch_module"],
    "objects": ["hash_file", "load_pickle", "save_pickle"],
    "paths": [
        "get_parent_path",
        "build_all_paths",
        "build_top_paths",
        "get_all_attributes",
        "list_all_paths",
    ],
    "processes": ["run_process"],
}


if TYPE_CHECKING:
    from pocketwelt.configs import BaseConfig
    from pocketwelt.encodings import b64_decode, b64_encode
    from pocketwelt.files import compress_file, decompress_file
    from pocketwelt.logs import (
        getCustomLogger,
        stdout_to_logger,
        CustomLogger,
        ColorFormatter,
    )
    from pocketwelt.modules import fetch_module
    from pocketwelt.objects import hash_file, load_pickle, save_pickle
    from pocketwelt.paths import (
        build_all_paths,
        build_top_paths,
        get_all_attributes,
        get_parent_path,
        list_all_paths,
    )
    from pocketwelt.processes import run_process

else:
    sys.modules[__name__] = LazyModule(
        module_name=__name__,
        module_file=globals()["__file__"],
        import_structure=_import_structure,
    )



import copy
import logging
import os
from contextlib import redirect_stdout
from datetime import datetime
from io import StringIO
from typing import Any, Callable, Literal, Optional, Union

from pocketwelt.colors import RGBColor

try:
    import click
except ImportError:
    print(
        "Package called `click` has not been found. "
        "Importing replacement class from `pocketwelt.colors`"
    )
    from pocketwelt.colors import click


DATETIME_FMT = "%Y-%m-%d %H:%M:%S"
LOG_FMT = "%(asctime)s  %(levelname)-8s %(filename)s:%(lineno)d >>  %(message)s"


class ColorFormatter(logging.Formatter):
    """
    Custom colored formatter converting logRecord to text.

    Adapted from:
        * https://github.com/encode/uvicorn/blob/master/uvicorn/logging.py
    """

    level_name_colors = {
        logging.DEBUG: lambda msg: click.style(str(msg), fg=RGBColor.TEAL),
        logging.INFO: lambda msg: click.style(str(msg), fg=RGBColor.GREEN),
        logging.WARNING: lambda msg: click.style(str(msg), fg=RGBColor.YELLOW),
        logging.ERROR: lambda msg: click.style(str(msg), fg=RGBColor.CRIMSON),
        logging.CRITICAL: lambda msg: click.style(str(msg), fg=RGBColor.RED),
    }

    def __init__(
        self,
        fmt: Optional[str] = LOG_FMT,
        datefmt: Optional[str] = DATETIME_FMT,
        style: Literal["%", "{", "$"] = "%",
        use_colors: bool = True,
    ) -> None:
        """
        Initialize the ColorFormatter.

        Args:
            fmt (Optional[str], optional): The format string. Defaults to `None`.
            datefmt (Optional[str], optional): The date format string. Defaults to `None`.
            style (Literal["%", "{", "$"], optional): The style of the format string. Defaults to '%'.
            use_colors (bool, optional): Whether to use colors in the output. Defaults to `False`.
        """
        self.use_colors = use_colors
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def color_message(self, level_no: int, msg: str) -> str:
        """
        Color the log message based on the log level.

        Args:
            level_no (int): The log level number.
            msg (str): The log message.

        Returns:
            str: The colored log message.
        """
        coloring_func = self.level_name_colors.get(
            level_no, lambda msg: click.style(str(msg), fg=RGBColor.GREEN)
        )

        return coloring_func(msg)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format the specified record.

        Args:
            record (logging.LogRecord): The log record to be formatted.

        Returns:
            str: The formatted log record as a string, possibly with color applied.
        """
        s = super().format(record)
        if self.use_colors:
            s = self.color_message(record.levelno, s)

        return s


class CustomLogger(logging.Logger):
    """
    Custom Logger class accepting formatter objects and adding file logging option.
    """

    def __init__(
        self,
        name: str,
        level: Union[str, int] = logging.WARNING,
        formatter: Optional[logging.Formatter] = None,
        log_directory: Optional[str] = None,
        use_colors: bool = True,
        **fmt_kwargs: Any,
    ) -> None:
        """
        Initialize the custom logger class which accepts `ColorFormatter` and log file to
        dump your logs to.

        Args:
            name (str): The name of the logger.
            level (Union[str, int]): The logging level for this logger.
            formatter (logging.Formatter, optional): Formatter to set for stream and file handlers.
                Defaults to `None` (use default formatters).
            log_directory (str, optional): The directory to store log files.
                If provided, log files will be created. Defaults to `None`.
            use_colors (bool): Whether to use colored logs. Defaults to `True`.
            **fmt_kwargs: Formatting keywords: `fmt` and `datefmt` for the formatter.
        """
        super().__init__(name, level)

        # Prepare formatter
        formatter = formatter or ColorFormatter(
            use_colors=use_colors,
            fmt=fmt_kwargs.get("fmt", LOG_FMT),
            datefmt=fmt_kwargs.get("datefmt", DATETIME_FMT),
        )
        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(formatter)
        self.addHandler(handler)

        # Create file logging
        if log_directory is not None:
            os.makedirs(log_directory, exist_ok=True)
            log_name = (
                "log"
                + f"_{name.lower()}_"
                + datetime.now().strftime("%Y-%m-%d_%H%M")
                + ".txt"
            )
            log_path = os.path.join(log_directory, log_name)
            handler = logging.FileHandler(log_path, mode="a")
            handler.setLevel(level)

            # Create non-colored formatter for file logs
            if hasattr(formatter, "use_colors"):
                file_formatter = copy.copy(formatter)
                setattr(file_formatter, "use_colors", False)
            handler.setFormatter(file_formatter)
            self.addHandler(handler)

    def setLevel(self, level: Union[str, int]) -> None:
        """
        Set the logging level for this logger and all its handlers.

        Args:
            level (Union[str, int]): The logging level to set. Can be a string
                (e.g., 'INFO', 'DEBUG') or an integer representing the level.
        """
        super().setLevel(level)
        for handler in self.handlers:
            handler.setLevel(level)


def stdout_to_logger(logger: logging.Logger, func: Callable, *args) -> str:
    """
    Redirect stdout to a logger and execute a function. Method can be used when you
    are not sure were given `print()` statements are located in huge codebases, so
    you can wrap any method in this and force logs via your logging system.

    Args:
        logger (logging.Logger): Your own logger to use for logging the stdout output.
        func (Callable): The function to execute with stdout redirection.
        *args: Variable length argument list to pass to the function.

    Returns:
        Any: The return value of the executed function.
    """
    f = StringIO()
    with redirect_stdout(f):
        res = func(*args)

    o = f.getvalue()
    for line in o.splitlines():
        logger.info(line)

    return res


# method name follows logging.py convention
def getCustomLogger(
    name: str, level: Union[str, int] = logging.WARNING
) -> CustomLogger:
    """
    Create custom logger instance.

    Args:
        name (str): Name of the logger.
        level (Union[str, int], optional): The logging level for the logger.
            Can be a string (e.g., 'INFO', 'DEBUG') or an integer. Defaults to logging.WARNING.

    Returns:
        CustomLogger:
            Instance of the `CustomLogger`.

    Note:
        This method temporarily replaces base class for loggers. We do this because
        we do not want to multiply the `Manager` object created initially in logging.py.
        Instead we use this object to create loggers with our custom class - logger hierarchy
        is intact and we get our colorful logs this way.
    """
    logging.Logger.manager.setLoggerClass(CustomLogger)
    _logger = logging.Logger.manager.getLogger(name)  # handles retrieving existing logger
    _logger.setLevel(level)
    logging.Logger.manager.setLoggerClass(logging.Logger)

    return _logger  # type: ignore



import hashlib
import os
import pickle
from typing import Any, Optional, Union


def save_pickle(obj: Any, path: str, replace: bool = False) -> None:
    """
    Save an object to a file using pickle serialization.

    Args:
        obj (Any): Python object to be saved.
        path (str): File path where the object will be saved.
        replace (bool, optional): If `True`, overwrites the file if it already exists.
            Defaults to `False`.

    Raises:
        FileExistsError: If the file already exists and `replace` is `False`.
        ValueError: If given path does not containt .pkl extension.
    """
    if os.path.exists(path) and not replace:
        raise FileExistsError(
            f"{path} already exists - either use `replace=True` or rename/move the file."
        )
    if not path.endswith(".pkl"):
        raise ValueError("Only pickle (.pkl) files are supported!")
    dir = os.path.dirname(path)
    if dir != "":
        os.makedirs(dir, exist_ok=True)
    with open(path, "wb") as f:
        pickle.dump(obj, f)


def load_pickle(path: str) -> Any:
    """
    Load an object from a file using pickle deserialization.

    Args:
        path (str): The file path from which to load the object.

    Raises:
        FileNotFoundError: If the specified file does not exist.
        ValueError: If given path does not containt .pkl extension.

    Returns:
        Any: Deserialized Python object loaded from the file.
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} does not exist!")
    if not path.endswith(".pkl"):
        raise ValueError("Only pickle (.pkl) files are supported!")

    return pickle.load(open(path, "rb"))


def hash_file(file: Union[str, bytes], size: Optional[int] = None) -> str:
    """
    Hash file using SHA-256.

    Args:
        file (Union[str, bytes]): The file to hash. Can be either
            a file path (str) or file content (bytes).
        size (int, optional): The number of bytes to read from the file.
            If `None`, reads the entire file. Defaults to `None`.

    Returns:
        str: The hexadecimal representation of the SHA-256 hash.

    Raises:
        AssertionError: If the filetype is neither str nor bytes.
    """
    assert isinstance(
        file, (str, bytes)
    ), f"Invalid filetype - must be `str` or `bytes`, got {type(file)}."

    if isinstance(file, str):
        with open(file, "rb") as f:
            file = f.read(size)

    return hashlib.sha256(file).hexdigest()



import base64
import os
from io import BytesIO, StringIO
from typing import Any, Optional, Union


MIME_HEADERS = {
    "png": "image/png",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "html": "text/html",
    "glb": "text/plain",
    "text": "text/plain",
    "wav": "audio/wav",
    "mp3": "audio/mpeg",
    "webm": "audio/webm",
    "mp4": "video/mp4",
    "mkv": "video/x-matroska",
    "avi": "video/x-msvideo",
    "zip": "application/zip",
    "md": "text/markdown",
    "pdf": "application/pdf",
}


def b64_encode(obj: Any, type_or_header: str) -> str:
    """
    Encode an object to a base64 string with a MIME header.

    Args:
        obj (Any): The object to encode. Can be a string (file path or content),
            `BytesIO`, `StringIO`, `bytes`, or any other object that can be
            converted to bytes. If `obj` does not support casting to bytes, will
            raise its own exception.
        type_or_header (str): Either a key from `MIME_HEADERS` or a full MIME header.

    Raises:
        ValueError: If the provided `type_or_header` is not supported or invalid.

    Returns:
        str: A string in the format "data:<mime_header>;base64,<encoded_data>".

    Examples:
        >>> b64_encode("hello.txt", "text")
        'data:text/plain;base64,aGVsbG8udHh0'

        >>> b64_encode(b"binary data", "application/octet-stream")
        'data:application/octet-stream;base64,YmluYXJ5IGRhdGE='
    """
    # First extract proper header
    if "/" not in type_or_header and type_or_header not in MIME_HEADERS:
        raise ValueError(
            f"{type_or_header} not supported. Either pass proper header"
            " or use provided headers (see `encodings.MIME_HEADERS`)."
        )
    mime_header = (
        type_or_header if "/" in type_or_header else MIME_HEADERS.get(type_or_header)
    )

    # For strings either enocde file content or text
    if isinstance(obj, str):
        if os.path.exists(obj):
            with open(obj, "rb") as file:
                enc_obj = base64.b64encode(file.read())
        else:
            enc_obj = base64.b64encode(obj.encode("utf-8"))

    # For buffers encode their value
    elif isinstance(obj, (BytesIO, StringIO)):
        obj.seek(0)  # make sure we read from the start
        if isinstance(obj, BytesIO):
            enc_obj = base64.b64encode(obj.getvalue())
        else:
            enc_obj = base64.b64encode(obj.getvalue().encode("utf-8"))
        obj.seek(0)  # return to starting position

    # For bytes simply encode them
    elif isinstance(obj, bytes):
        enc_obj = base64.b64encode(obj)

    # Anything else try casting to bytes and encoding
    else:
        enc_obj = base64.b64encode(bytes(obj))

    return f"data:{mime_header};base64,{enc_obj.decode('utf-8')}"


def b64_decode(encoded_obj: str, filepath: Optional[str] = None) -> Union[str, BytesIO]:
    """
    Decode a base64 encoded string.

    Args:
        encoded_obj (str): The base64 encoded string to decode.
        filepath (str, optional): The path to save the decoded content as a file.
            If `None`, the decoded content is returned as a `BytesIO` object. Defaults to `None`.

    Raises:
        FileExistsError: If the specified file path already exists.

    Returns:
        Union[str, BytesIO]: If `fileobj` is a valid path, returns the absolute path of the
            created file as a string. If `fileobj` is `None`, returns a `BytesIO` object containing
            the decoded content.

    Examples:
        >>> decoded = b64_decode("data:text/plain;base64,aGVsbG8=")
        >>> decoded.getvalue()
        b'hello'

        >>> file_path = b64_decode("data:text/plain;base64,aGVsbG8=", "output.txt")
        >>> with open(file_path, "r") as f:
        ...     print(f.read())
        hello
    """

    # Decode object
    decoded_obj = base64.b64decode(encoded_obj.split(",")[-1])

    # If not string, dump to binary buffer
    if filepath is None:
        return BytesIO(decoded_obj)

    # If string, treat as path
    if isinstance(filepath, str):
        if os.path.exists(filepath):
            raise FileExistsError(f"{filepath} is taken - interrupting decoding.")
        _dir = os.path.dirname(filepath)
        if _dir != "":
            os.makedirs(_dir, exist_ok=True)
        with open(filepath, "wb") as f:
            f.write(decoded_obj)
        return os.path.abspath(filepath)



import subprocess


def run_process(command: str) -> str:
    """
    Run a shell command and return its output.

    Args:
        command (str): The shell command to execute.

    Raises:
        subprocess.SubprocessError: If the command execution results in an error.

    Returns:
        str: Output of the executed command.
    """
    result = subprocess.run(command.split(), capture_output=True, text=True)
    if result.stderr:
        raise subprocess.SubprocessError(result.stderr)

    return result.stdout



import importlib
from importlib.util import find_spec
import os
from types import ModuleType
from typing import Any, Dict, List

from pocketwelt.paths import get_all_attributes, build_all_paths


def is_package_available(name: str) -> bool:
    """
    Use to check for installation of a package. Finding path of the package
    does not import it, thus we can notify the user if some packages are missing.

    Args:
        name (str): Name of the package to check.
    """
    return find_spec(name) is not None


class LazyModule(ModuleType):
    """
    Module for lazy importing.

    Heavily based on:
        * https://github.com/optuna/optuna/blob/master/optuna/integration/__init__.py
        * https://github.com/huggingface/diffusers/blob/main/src/diffusers/utils/import_utils.py
    """

    def __init__(
        self, module_name: str, module_file: str, import_structure: Dict[str, List[str]]
    ):
        """
        Initialize a LazyModule instance.

        Args:
            module_name (str): Name of the module.
            module_file (str): File path of the module.
            import_structure (Dict[str, List[str]]): A dictionary representing the import structure.
        """
        super().__init__(module_name)
        self.__file__ = module_file
        self.__all__ = get_all_attributes(import_structure)
        self.__path__ = [os.path.dirname(module_file)]

        # Attributes for lazy importing based on structure
        self._import_structure = import_structure
        self._modules = set(import_structure.keys())
        self._objects_to_modules = build_all_paths(import_structure)

    def __getattr__(self, name: str) -> Any:
        """
        Retrieve module when requested from the namespace.

        Args:
            name (str): The name of the attribute to retrieve.

        Raises:
            AttributeError: If the attribute does not exist in the module.

        Returns:
            Any: The requested module or attribute.

        """
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._objects_to_modules.keys():
            # Get the module and then create the whole import path
            _modules = self._objects_to_modules[name]
            module = self._get_module(".".join(_modules))
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)

        return value

    def _get_module(self, module_name: str):
        """
        Import the given module.

        Args:
            module_name (str): The name of the module to import.

        Raises:
            RuntimeError: If the module cannot be imported.

        Returns:
            module: The imported module.
        """
        try:
            return importlib.import_module("." + module_name, self.__name__)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import {self.__name__}.{module_name} because of the following error:\n\n{e}"
            ) from e

    def __reduce__(self):
        """
        This method is used to ensure that the LazyModule objectcan be pickled and unpickled correctly.
        """
        return (self.__class__, (self.__name__, self.__file__, self._import_structure))



import gzip
import pickle
import shutil
from typing import Any


def compress_file(file_path: str, compressed_file_path: str) -> None:
    """
    Compress file using GZIP.

    Args:
        file_path (str): The path to the file to be compressed.
        compressed_file_path (str): The path where the compressed file will be saved.
    """
    with open(file_path, "rb") as f_in:
        with open(compressed_file_path, "wb") as f_out:
            with gzip.GzipFile(file_path, "wb", fileobj=f_out) as f_out:
                shutil.copyfileobj(f_in, f_out)


def decompress_file(compressed_file_path: str) -> Any:
    """
    Decompress a GZIP-compressed file and return its contents.

    Args:
        compressed_file_path (str): The path to the compressed file.

    Returns:
        Any: The decompressed and deserialized content of the file.
            We first read the content and then try unpickling. If that fails
            we simply decode the content from bytes.
    """
    with gzip.open(compressed_file_path, "rb") as f_in:
        content = f_in.read()
        try:
            return pickle.loads(content)
        except pickle.UnpicklingError:
            return content.decode("utf-8")



from types import SimpleNamespace
from typing import Any, Dict, get_origin


class BaseConfig(SimpleNamespace):
    """
    Simple configuration class designed to work as pydantic BaseModel.

    Example:
        .. code-block:: python

            class ExampleConfig(BaseConfig):
                host: str
                port: int
                use_ssl: bool = False

            # Example usage
            example_config = ExampleConfig(host="localhost", port=8080)
            print(example_config.host)  # Output: localhost
            print(example_config.port)  # Output: 8080
            print(example_config.use_ssl)  # Output: False
    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the BaseConfig instance.

        Args:
            **kwargs: Arbitrary keyword arguments representing configuration variables.

        Raises:
            AttributeError: If a provided keyword argument does not match any attribute in the class.
            TypeError: If a provided keyword argument does not match the expected type.
            ValueError: If any required arguments are missing.
        """
        # Very primitive way of disabling the direct construction
        assert (
            self.__class__.__base__ != SimpleNamespace
        ), "BaseConfig cannot be initiated directly - can only be used for class inheritance."

        # Construct the config
        _required = self._get_empty_attrs()
        for k, v in kwargs.items():
            # Check if attribute exists, retrieve type
            if not hasattr(self, k) and k not in self.__annotations__:
                raise AttributeError(
                    f"`{self.__class__.__name__}` does not contain config variable named `{k}`."
                )
            _type = self.__annotations__.get(k, None)

            # Check generic type origin to avoid TypeError
            if hasattr(_type, "__origin__"):
                _type = get_origin(_type)

            # Validate type & set attribute
            if _type and not isinstance(v, _type):
                raise TypeError(
                    f"`{k}` should be of type `{_type.__name__}`, but is of type`{type(v).__name__}`."
                )
            setattr(self, k, v)
            if k in _required:
                _required.remove(k)

        if len(_required) > 0:
            raise ValueError(
                f"The following arguments were missing: {','.join(_required)}"
            )

    def _get_empty_attrs(self):
        """
        Get a list of attributes that are annotated but not set.

        Returns:
            List[str]: A list of attribute names that are annotated but not set.
        """
        _empty_attrs = []
        for attr in self.__annotations__:
            if not hasattr(self, attr):
                _empty_attrs.append(attr)

        return _empty_attrs

    def to_dict(self) -> Dict[str, Any]:
        """
        Save the configuration to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the configuration.
        """
        _dict = {}
        for attr in self.__annotations__:
            _dict[attr] = getattr(self, attr, None)

        return _dict

    def to_txt(self, save_path: str) -> None:
        """
        Save the current config to a text file.

        Args:
            save_path (str): Path where the configuration text file will be saved.
        """
        with open(save_path, "w+") as f:
            for k, v in self.to_dict().items():
                f.write(f"{k}: {v}\n")



import importlib
from typing import Optional

from pocketwelt.processes import run_process


def fetch_module(package: str, alias: Optional[str] = None) -> None:
    """
    Try importing a package and install it if it's not available.

    Args:
        package (str): Name of the package to import and potentially install.
        alias (str): Alias for module used in code (e.g. `np` is an alias for `numpy`).

    Raises:
        ImportError: If the package cannot be imported after installation attempt.

    Note:
        This function modifies the global namespace by adding the imported module.

    Example:
        >>> fetch_module('numpy')
        # If numpy is not installed, it will be installed and then imported.
        # If numpy is already installed, it will just be imported.
    """
    try:
        importlib.import_module(package)
    except (ImportError, ModuleNotFoundError):
        run_process(f"python -m pip install {package}")
    finally:
        globals()[alias or package] = importlib.import_module(package)



import os
from typing import Any, Dict, List, Optional


def get_parent_path(path: str, level: int = 1, absolute: bool = True) -> str:
    """
    Find parent of a given path.

    Args:
        path (str): Path to get parent of.
        level (int, optional): How distant parent to look for. Defaults to 1.
        absolute (str): Return the absolute path to the parent. Defaults to `True`.

    Returns:
        str: Parent path of the given path.
    """
    assert level >= 0, "`level` must be at least 0."

    if absolute:
        path = os.path.abspath(path)
    if level > 0:
        return get_parent_path(os.path.dirname(path), level - 1, False)

    return path


def list_all_paths(directory: str) -> List[str]:
    """
    List all paths from the given directory.

    Args:
        directory (str): Path to the directory.

    Returns:
        List[str]: List of all paths in the directory.
    """
    paths = []
    for root, dirs, files in os.walk(directory):
        for name in dirs + files:
            paths.append(os.path.join(root, name))

    return paths


def build_all_paths(
    structure: Dict[str, Any],
    current_path: Optional[list] = None,
    result: Optional[dict] = None,
) -> Dict[str, Any]:
    """
    Recursively build a dictionary of all paths from a nested dictionary.

    Args:
        structure (Dict[str, Any]): The nested dictionary structure to process.
        current_path (Optional[list], optional): The current path being processed.
            Defaults to `None`.
        result (Optional[dict], optional): The dictionary to store the results.
            Defaults to `None`.

    Returns:
        Dict[str, Any]: A dictionary containing all paths from the nested dictionary.
    """
    if result is None:
        result = dict()
    if current_path is None:
        current_path = list()

    for key, value in structure.items():
        if isinstance(value, dict):
            build_all_paths(value, current_path + [key], result)
        else:
            result[key] = current_path
            for item in value:
                result[item] = current_path + [key]

    return result


def build_top_paths(
    structure: Dict[str, Any], current_path: Optional[str] = None
) -> List[str]:
    """
    Recursively build a list of all possible paths for a given nested dictionary.

    Args:
        structure (Dict[str, Any]): The nested dictionary structure to process.
        current_path (Optional[str], optional): The current path being processed.
            Defaults to `None`.

    Returns:
        List[str]: A list of all possible paths from the nested dictionary.
    """
    paths = []
    for key, value in structure.items():
        if isinstance(value, dict):
            for sub_path in build_top_paths(value, key):
                paths.append(f"{sub_path}")
        else:
            for item in value:
                paths.append(
                    f"{key}.{item}"
                    if current_path is None
                    else f"{current_path}.{key}.{item}"
                )

    return paths


def get_all_attributes(
    structure: Dict[str, Any], attrs: Optional[List[str]] = None
) -> List[str]:
    """
    Recursively retrieve all names that exist in the nested dictionary.

    Args:
        structure (Dict[str, Any]): The nested dictionary structure to process.
        attrs (Optional[List[str]], optional): The list to store the attribute names.
            Defaults to `None`.

    Returns:
        List[str]: A list of all unique attribute names in the nested dictionary.
    """
    if attrs is None:
        attrs = []

    for key, value in structure.items():
        attrs.append(key)
        if isinstance(value, dict):
            get_all_attributes(value, attrs)
        elif isinstance(value, list):
            for item in value:
                attrs.append(item)

    return list(set(attrs))
